DP_CHECKPOINT_FILE = 'model.ckpt'
DP_DISP_FILE = 'lcurve.out'
DP_PROFILING_FILE = 'timeline.json'
DP_INPUT_FILE = 'input.json'
DP_FROZEN_MODEL = 'frozen_model.pb'
DP_ORIGINAL_MODEL = 'original_model.pb'

MODEL_DEVI_OUT = 'model_devi.out'
MODEL_DEVI_NEU_OUT = 'model_devi_neu.out'
MODEL_DEVI_RED_OUT = 'model_devi_red.out'

LAMMPS_DUMP_DIR = 'traj'
LAMMPS_DUMP_SUFFIX = '.lammpstrj'

SELECTOR_OUTPUT = 'selector_output'


DEFAULT_LASP_IN = {
    "Run_type": 15,
    "SSW.SSWsteps": 20,
    "SSW.Temp": 300,
    "SSW.NG": 10,
    "SSW.NG_cell": 8,
    "SSW.ds_atom": 0.6,
    "SSW.ds_cell": 0.5,
    "SSW.ftol": 0.05,
    "SSW.strtol": 0.05,
    "SSW.MaxOptstep": 300,
    "SSW.printevery": "T",
    "SSW.printselect": 0,
    "SSW.printdelay": -1,
    "SSW.output": "F"
}


DEFAULT_LAMMPS_TEMPLATE_FOR_DP_SSW = """\
units           metal
boundary        p p p
atom_style      atomic
atom_modify map yes

$$read_data_section

$$force_field_section

compute peratom all pressure NULL virial
"""


DEFAULT_ASAP_ASCF_DESC = {
    'preset': None,
    # those params following the convention of dscribe
    # https://singroup.github.io/dscribe/latest/tutorials/descriptors/acsf.html
    'r_cut': 3.5,

    'reducer_type': 'average',
    'element_wise': False,
    'zeta': 1,
}


DEFAULT_ASAP_SOAP_DESC = {
    'preset': None,

    # those params following the convention of dscribe
    # https://singroup.github.io/dscribe/latest/tutorials/descriptors/soap.html
    'r_cut': 3.5,
    'n_max': 6,
    'l_max': 6,
    'sigma': 0.5,

    'crossover': False,
    'rbf': 'gto',

    'reducer_type': 'average',
    'element_wise': False,
    'zeta': 1,
}

DEFAULT_ASAP_PCA_REDUCER = {
    'type': 'PCA',
    'parameter': {
        'n_components': 3,
        'scalecenter': True,
    }
}

# LAMMPS

_DEFAULT_LAMMPS_TOP = '''\
$$VARIABLES
$$EXTRA_VARS

$$INITIALIZE

$$POST_INIT

$$READ_DATA
$$POST_READ_DATA

$$MASS_MAP

$$SET_ATOM_TYPE
'''

_DEFAULT_LAMMPS_BOTTOM = '''\
$$POST_FORCE_FIELD

$$SIMULATION
$$POST_SIMULATION

$$RUN
$$POST_RUN
'''

_DP_FORCE_FIELD = '''\
pair_style deepmd $$DP_MODELS out_freq ${THERMO_FREQ} out_file model_devi.out
pair_coeff * *
'''

_DP_FEP_DUAL_FORCE_FIELD = '''\
variable LAMBDA_i equal 1-v_LAMBDA_f

pair_style  hybrid/overlay &
            deepmd $$DP_NEU_MODELS out_freq ${THERMO_FREQ} out_file model_devi_neu.out &
            deepmd $$DP_RED_MODELS out_freq ${THERMO_FREQ} out_file model_devi_red.out
pair_coeff  * * deepmd 1 *
pair_coeff  * * deepmd 2 *

fix PES_Sampling all adapt 0 &
    pair deepmd:1 scale * * v_LAMBDA_f &
    pair deepmd:2 scale * * v_LAMBDA_i
'''

_DP_FEP_UNI_FORCE_FIELD = '''\
variable LAMBDA_i equal 1-v_LAMBDA_f

pair_style  hybrid/overlay &
            $$PAIR_STYLE_EXT &
            deepmd $$DP_MODELS_0 type_order $$DP_FEP_INI_TYPE_ORDER &
            deepmd $$DP_MODELS_0 type_order $$DP_FEP_FIN_TYPE_ORDER
$$PAIR_COEFF_EXT
pair_coeff  * * deepmd 1 *
pair_coeff  * * deepmd 2 *

fix PES_Sampling all adapt 0 &
    pair deepmd:1 scale * * v_LAMBDA_f &
    pair deepmd:2 scale * * v_LAMBDA_i
'''

def _get_fep_rerun_setting(ns: str, in_data:str, in_traj: str):
    return f'''\
clear
# Rerun model deviation: {ns}
shell mkdir traj-{ns}
$$INITIALIZE
read_data {in_data}
$$MASS_MAP_BASE
$$POST_READ_DATA

pair_style deepmd $$DP_MODELS out_freq 1 out_file model_devi_{ns}.out type_order $$DP_DEFAULT_TYPE_ORDER
pair_coeff * *

thermo 1
thermo_style custom step temp pe ke etotal
thermo_modify format float %15.7f
dump 1 all custom 1 traj-{ns}/*.lammpstrj id type x y z
rerun {in_traj} first 0 last 1000000000000 every 1 dump x y z box yes
'''

# Notes
# 1. LAMMPS shell command use triple quote to pass a string with space and special chars in it.
# e.g """ "$$SPECORDER" """
# 2. ase read lammps-data format need to set style to "atomic"
# ref: https://gitlab.com/ase/ase/-/issues/1126
_FEP_UNI_BOTTOM = '\n'.join(['''\
# Run post processing script
shell env > debug.env.txt
shell cat traj/*.lammpstrj > traj.lammpstrj
shell cp ${DATA_FILE} ini.data
clear
shell $$AI2KIT_CMD tool ase read traj.lammpstrj --format lammps-dump-text --specorder """ "$$SPECORDER" """ - write ini.lammpstrj --format lammps-dump-text  --type_map """ "$$SPECORDER_BASE" """
shell $$AI2KIT_CMD tool ase read traj.lammpstrj --format lammps-dump-text --specorder """ "$$SPECORDER" """ - delete_atoms """ "$$DELETE_ATOMS" """ - write fin.lammpstrj --format lammps-dump-text --type_map """ "$$SPECORDER_BASE" """
shell $$AI2KIT_CMD tool ase read ini.data --format lammps-data --style "atomic" - delete_atoms """ "$$DELETE_ATOMS" """ - write fin.data --format lammps-data
''',
    _get_fep_rerun_setting('ini', 'ini.data', 'ini.lammpstrj'),
    _get_fep_rerun_setting('fin', 'fin.data', 'fin.lammpstrj'),
])


PRESET_LAMMPS_INPUT_TEMPLATE = {
    'default': '\n'.join([ _DEFAULT_LAMMPS_TOP, _DP_FORCE_FIELD, _DEFAULT_LAMMPS_BOTTOM]),
    # 2 models fep
    'fep-2m': '\n'.join([ _DEFAULT_LAMMPS_TOP, _DP_FEP_DUAL_FORCE_FIELD, _DEFAULT_LAMMPS_BOTTOM]),
    # 1 model fep
    'fep': '\n'.join([ _DEFAULT_LAMMPS_TOP, _DP_FEP_UNI_FORCE_FIELD, _DEFAULT_LAMMPS_BOTTOM, _FEP_UNI_BOTTOM]),
}



from ai2_kit.core.executor import BaseExecutorConfig
from ai2_kit.core.artifact import ArtifactMap
from ai2_kit.core.log import get_logger
from ai2_kit.core.util import load_yaml_files
from ai2_kit.core.resource_manager import ResourceManager
from ai2_kit.domain import (
    deepmd,
    iface,
    lammps,
    selector,
    cp2k,
    constant as const,
    updater,
)
from ai2_kit.core.checkpoint import set_checkpoint_file, apply_checkpoint

from pydantic import BaseModel
from typing import Dict, List, Optional, Any
from fire import Fire

import asyncio
import copy
import itertools
import os

logger = get_logger(__name__)


class FepExecutorConfig(BaseExecutorConfig):
    class Context(BaseModel):

        deepmd: deepmd.CllDeepmdContextConfig
        lammps: lammps.CllLammpsContextConfig
        cp2k: cp2k.CllCp2kContextConfig

    context: Context


class WorkflowConfig(BaseModel):
    class General(BaseModel):
        type_map: List[str]
        mass_map: List[float]
        max_iters: int = 10

    class Branch(BaseModel):
        deepmd: deepmd.CllDeepmdInputConfig
        cp2k: cp2k.CllCp2kInputConfig
        threshold: selector.CllModelDeviSelectorInputConfig

    class Update(BaseModel):
        walkthrough: updater.CllWalkthroughUpdaterInputConfig

    general: General
    neu: Branch
    red: Branch
    lammps: lammps.CllLammpsInputConfig
    update: Update


class FepWorkflowConfig(BaseModel):
    executors: Dict[str, FepExecutorConfig]
    artifacts: ArtifactMap
    workflow: Any


def run_workflow(*config_files, executor: Optional[str] = None,
                 path_prefix: Optional[str] = None, checkpoint: Optional[str] = None):
    """
    Training ML potential for FEP
    """
    if checkpoint is not None:
        set_checkpoint_file(checkpoint)

    config_data = load_yaml_files(*config_files)
    config = FepWorkflowConfig.parse_obj(config_data)

    if executor not in config.executors:
        raise ValueError(f'executor {executor} is not found')
    if path_prefix is None:
        raise ValueError('path_prefix should not be empty')

    iface.init_artifacts(config.artifacts)
    resource_manager = ResourceManager(
        executor_configs=config.executors,
        artifacts=config.artifacts,
        default_executor=executor,
    )
    return asyncio.run(cll_mlp_training_workflow(config, resource_manager, executor, path_prefix))


async def cll_mlp_training_workflow(config: FepWorkflowConfig, resource_manager: ResourceManager, executor: str, path_prefix: str):
    context_config = config.executors[executor].context
    raw_workflow_config = copy.deepcopy(config.workflow)

    # output of each step
    neu_label_output: Optional[iface.ICllLabelOutput] = None
    red_label_output: Optional[iface.ICllLabelOutput] = None

    neu_selector_output: Optional[iface.ICllSelectorOutput] = None
    red_selector_output: Optional[iface.ICllSelectorOutput] = None

    neu_train_output: Optional[iface.ICllTrainOutput] = None
    red_train_output: Optional[iface.ICllTrainOutput] = None

    explore_output: Optional[iface.ICllExploreOutput] = None

    # cursor of update table
    update_cursor = 0
    # Start iteration
    for i in itertools.count(0):

        # parse workflow config
        workflow_config = WorkflowConfig.parse_obj(raw_workflow_config)
        if i >= workflow_config.general.max_iters:
            logger.info(f'Iteration {i} exceeds max_iters, stop iteration.')
            break

        # shortcut for type_map and mass_map
        type_map = workflow_config.general.type_map
        mass_map = workflow_config.general.mass_map

        # decide path prefix for each iteration
        iter_path_prefix = os.path.join(path_prefix, f'iters-{i:03d}')
        # prefix of checkpoint
        cp_prefix = f'iters-{i:03d}'

        # label: cp2k
        red_cp2k_input = cp2k.CllCp2kInput(
            config=workflow_config.red.cp2k,
            type_map=type_map,
            system_files=[] if red_selector_output is None else red_selector_output.get_model_devi_dataset(),
            initiated=i > 0,
        )
        red_cpk2_context = cp2k.CllCp2kContext(
            config=context_config.cp2k,
            path_prefix=os.path.join(iter_path_prefix, 'red-label-cp2k'),
            resource_manager=resource_manager,
        )

        neu_cp2k_input = cp2k.CllCp2kInput(
            config=workflow_config.neu.cp2k,
            type_map=type_map,
            system_files=[] if neu_selector_output is None else neu_selector_output.get_model_devi_dataset(),
            initiated=i > 0,
        )
        neu_cp2k_context = cp2k.CllCp2kContext(
            config=context_config.cp2k,
            path_prefix=os.path.join(iter_path_prefix, 'neu-label-cp2k'),
            resource_manager=resource_manager,
        )

        red_label_output, neu_label_output = await asyncio.gather(
            apply_checkpoint(f'{cp_prefix}/cp2k/red')(cp2k.cll_cp2k)(red_cp2k_input, red_cpk2_context),
            apply_checkpoint(f'{cp_prefix}/cp2k/neu')(cp2k.cll_cp2k)(neu_cp2k_input, neu_cp2k_context),
        )

        # Train
        red_deepmd_input = deepmd.CllDeepmdInput(
            config=workflow_config.red.deepmd,
            type_map=type_map,
            old_dataset=[] if red_train_output is None else red_train_output.get_training_dataset(),
            new_dataset=red_label_output.get_labeled_system_dataset(),
            initiated=i > 0,
        )
        red_deepmd_context = deepmd.CllDeepmdContext(
            path_prefix=os.path.join(iter_path_prefix, 'red-train-deepmd'),
            config=context_config.deepmd,
            resource_manager=resource_manager,
        )
        neu_deepmd_input = deepmd.CllDeepmdInput(
            config=workflow_config.neu.deepmd,
            type_map=type_map,
            old_dataset=[] if neu_train_output is None else neu_train_output.get_training_dataset(),
            new_dataset=neu_label_output.get_labeled_system_dataset(),
            initiated=i > 0,
        )
        neu_deepmd_context = deepmd.CllDeepmdContext(
            path_prefix=os.path.join(iter_path_prefix, 'neu-train-deepmd'),
            config=context_config.deepmd,
            resource_manager=resource_manager,
        )

        red_train_output, neu_train_output = await asyncio.gather(
            apply_checkpoint(f'{cp_prefix}/deepmd/red')(deepmd.cll_deepmd)(red_deepmd_input, red_deepmd_context),
            apply_checkpoint(f'{cp_prefix}/deepmd/neu')(deepmd.cll_deepmd)(neu_deepmd_input, neu_deepmd_context),
        )

        # explore
        lammps_input = lammps.CllLammpsInput(
            config=workflow_config.lammps,
            type_map=type_map,
            mass_map=mass_map,
            dp_models={
                'NEU': neu_train_output.get_mlp_models(),
                'RED': red_train_output.get_mlp_models(),
            },
            preset_template='fep-2m'
        )
        lammps_context = lammps.CllLammpsContext(
            path_prefix=os.path.join(iter_path_prefix, 'explore-lammps'),
            config=context_config.lammps,
            resource_manager=resource_manager,
        )
        explore_output = await apply_checkpoint(f'{cp_prefix}/lammps')(lammps.cll_lammps)(lammps_input, lammps_context)

        # select
        red_selector_input = selector.CllModelDeviSelectorInput(
            config=workflow_config.red.threshold,
            model_devi_data=explore_output.get_model_devi_dataset(),
            model_devi_file=const.MODEL_DEVI_RED_OUT,
            type_map=type_map,
        )
        red_selector_context = selector.CllModelDevSelectorContext(
            path_prefix=os.path.join(
                iter_path_prefix, 'red-selector-threshold'),
            resource_manager=resource_manager,
        )

        neu_selector_input = selector.CllModelDeviSelectorInput(
            config=workflow_config.neu.threshold,
            model_devi_data=explore_output.get_model_devi_dataset(),
            model_devi_file=const.MODEL_DEVI_NEU_OUT,
            type_map=type_map,
        )
        neu_selector_context = selector.CllModelDevSelectorContext(
            path_prefix=os.path.join(iter_path_prefix, 'neu-selector-threshold'),
            resource_manager=resource_manager,
        )

        red_selector_output, neu_selector_output = await asyncio.gather(
            apply_checkpoint(f'{cp_prefix}/selector/red')(selector.cll_model_devi_selector)(red_selector_input, red_selector_context),
            apply_checkpoint(f'{cp_prefix}/selector/neu')(selector.cll_model_devi_selector)(neu_selector_input, neu_selector_context),
        )

        # Update
        update_config = workflow_config.update.walkthrough

        # nothing to update because the table is empty
        if not update_config.table:
            continue
        # keep using the latest config when it reach the end of table
        if update_cursor >= len(update_config.table):
            continue
        # update config
        update_cursor += 1


if __name__ == '__main__':
    # use python-fire to parse command line arguments
    Fire(run_workflow)



from ai2_kit.core.artifact import Artifact
from ai2_kit.core.script import BashTemplate
from ai2_kit.core.script import BashScript, BashStep
from ai2_kit.core.job import JobFuture, gather_jobs
from ai2_kit.core.log import get_logger
from ai2_kit.core.util import dict_nested_get

from pydantic import BaseModel
from typing import List
from dataclasses import dataclass
import os
import copy
import random
import sys
import json

from .iface import ICllTrainOutput, BaseCllContext
from .data import DataFormat, get_data_format, convert_to_deepmd_npy

from .constant import (
    DP_CHECKPOINT_FILE,
    DP_DISP_FILE,
    DP_PROFILING_FILE,
    DP_INPUT_FILE,
    DP_FROZEN_MODEL,
    DP_ORIGINAL_MODEL
)

logger = get_logger(__name__)


class CllDeepmdInputConfig(BaseModel):
    model_num: int = 4
    init_dataset: List[str]
    input_template: dict
    compress_model: bool = False

class CllDeepmdContextConfig(BaseModel):
    script_template: BashTemplate
    dp_cmd: str = 'dp'


@dataclass
class CllDeepmdInput:
    config: CllDeepmdInputConfig
    type_map: List[str]
    old_dataset: List[Artifact]  # training data used by previous iteration
    new_dataset: List[Artifact]  # training data used by current iteration
    initiated: bool = False  # FIXME: this seems to be a bad design idea

    def update_training_dataset(self, dataset: List[Artifact]):
        self.old_dataset += self.new_dataset
        self.new_dataset = dataset


@dataclass
class CllDeepmdContext(BaseCllContext):
    config: CllDeepmdContextConfig


@dataclass
class GenericDeepmdOutput(ICllTrainOutput):
    outputs: List[Artifact]
    input: CllDeepmdInput

    def get_mlp_models(self) -> List[Artifact]:
        return [a.join(DP_FROZEN_MODEL) for a in self.outputs]

    def get_training_dataset(self) -> List[Artifact]:
        return self.input.new_dataset + self.input.old_dataset


async def cll_deepmd(input: CllDeepmdInput, ctx: CllDeepmdContext):
    executor = ctx.resource_manager.default_executor

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    [converted_data_dir, tasks_dir] = executor.setup_workspace(
        work_dir, ['converted_input_data', 'tasks'])

    # initialization
    if not input.initiated:
        input.new_dataset += ctx.resource_manager.resolve_artifacts(
            input.config.init_dataset)

    # convert data type if necessary, only needed for new data as old data is already converted
    deepmd_npy_data: List[Artifact] = []
    data_to_be_converted: List[Artifact] = []

    for artifact in input.new_dataset:
        data = artifact.to_dict()
        format = get_data_format(data)  # type: ignore
        if format == DataFormat.DEEPMD_NPY or not format:
            deepmd_npy_data.append(artifact)
        else:
            data_to_be_converted.append(artifact)

    # convert data to deepmd/npy format
    converted_data_dirs = executor.run_python_fn(convert_to_deepmd_npy)(
        dataset=[a.to_dict() for a in data_to_be_converted],
        base_dir=converted_data_dir,
        type_map=input.type_map,
    )

    deepmd_npy_data += [Artifact.of(
        url=a['url'], format=a['format'], attrs=a['attrs']
    ) for a in converted_data_dirs]

    input.new_dataset = deepmd_npy_data

    jobs: List[JobFuture] = []
    output_dirs = []

    # train multiple models at the same time
    for i in range(input.config.model_num):
        # each model should be trained in its own task_dir
        task_dir = os.path.join(tasks_dir, str(i).zfill(3))
        executor.mkdir(task_dir)

        # create dp train input file
        # NOTE: dp v1 format is supported currently
        # TODO: support more params if it is necessary

        # ref: https://github.com/deepmodeling/dpgen2/blob/master/examples/ch4/param_CH4_deepmd-kit-2.1.1.json
        # ref: https://github.com/deepmodeling/dpgen2/blob/master/dpgen2/op/prep_dp_train.py
        # ref: https://github.com/deepmodeling/dpgen2/blob/master/dpgen2/op/run_dp_train.py

        dp_input = copy.deepcopy(input.config.input_template)
        training: dict = dp_input['training']

        # set output files
        training['disp_file'] = DP_DISP_FILE
        training['save_ckpt'] = DP_CHECKPOINT_FILE
        training['profiling_file'] = DP_PROFILING_FILE

        # set random seed
        discriptor = dp_input['model']['descriptor']
        if discriptor['type'] == 'hybrid':
            for d in discriptor['list']:
                d['seed'] = _random_seed()
        else:
            discriptor['seed'] = _random_seed()
        dp_input['model']['fitting_net']['seed'] = _random_seed()
        dp_input['training']['seed'] = _random_seed()

        # set training data
        VALIDATION_DATA_KEY = 'validation_data'
        train_systems = [a.url for a in input.old_dataset + input.new_dataset
                         if not dict_nested_get(a.attrs, ['deepmd', VALIDATION_DATA_KEY], False)]
        validation_systems = [a.url for a in input.old_dataset + input.new_dataset
                            if dict_nested_get(a.attrs, ['deepmd', VALIDATION_DATA_KEY], False)]

        training['systems'] = train_systems
        set_prefix: str = training.setdefault(
            'set_prefix', 'set')  # respect user input
        auto_prob_str = "prob_sys_size"
        training.setdefault('batch_size', 'auto')
        training['auto_prob_style'] = auto_prob_str

        # v2 training data
        training_data = {
            'systems': training['systems'],
            'set_prefix': training['set_prefix'],
            'auto_prob_style': training['auto_prob_style'],
            'batch_size': training['batch_size'],
        }
        training['training_data'] = training_data

        # ignore validation section if no data is provided, or else dp will throw error
        # OSError: [Errno cannot find valid a data system] Please check your setting for data systems
        if len(validation_systems) > 0:
            validation_data = {
                'systems': validation_systems,
                'set_prefix': training['set_prefix'],
                'batch_size': training['batch_size'],
            }
            training['validation_data'] = validation_data

        # other params
        dp_input['model']['type_map'] = input.type_map

        # write config to executor
        dp_input_text = json.dumps(dp_input, indent=2)
        dp_input_path = os.path.join(task_dir, DP_INPUT_FILE)
        executor.dump_text(dp_input_text, dp_input_path)

        # build script
        dp_cmd = ctx.config.dp_cmd
        dp_train_cmd = [dp_cmd, 'train', DP_INPUT_FILE]

        steps = [
            BashStep(cmd=dp_train_cmd, checkpoint='dp-train') # type: ignore
        ]

        if input.config.compress_model:
            dp_freeze_cmd = [dp_cmd, 'freeze', '-o', DP_ORIGINAL_MODEL]
            dp_compress_cmd = \
                [dp_cmd, 'compress', '-i', DP_ORIGINAL_MODEL, '-o', DP_FROZEN_MODEL]
            steps.append(BashStep(cmd=dp_freeze_cmd)) # type: ignore
            steps.append(BashStep(cmd=dp_compress_cmd)) # type: ignore
        else:
            dp_freeze_cmd = [dp_cmd, 'freeze', '-o', DP_FROZEN_MODEL]
            steps.append(BashStep(cmd=dp_freeze_cmd)) # type: ignore

        dp_train_script = BashScript(
            template=ctx.config.script_template,
            steps=steps # type: ignore
        )
        output_dirs.append(task_dir)

        # submit job
        job = executor.submit(dp_train_script.render(), cwd=task_dir, checkpoint_key=f'queue-job:dp-train:{task_dir}:{i}')
        jobs.append(job)

    await gather_jobs(jobs, max_tries=2)

    return GenericDeepmdOutput(
        input=input,
        outputs=[Artifact.of(
            url=url,
            format=DataFormat.DEEPMD_OUTPUT_DIR,
        ) for url in output_dirs]
    )


def _random_seed():
    return random.randrange(sys.maxsize) % (1 << 32)



from ai2_kit.core.script import BashTemplate, BashStep, BashScript
from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.log import get_logger
from ai2_kit.core.job import gather_jobs
from ai2_kit.core.util import list_split, dict_nested_get, dump_json, dump_text

from typing import List, Literal, Optional, Mapping, Sequence, Any
from pydantic import BaseModel, validator, root_validator
from dataclasses import dataclass
from string import Template
from allpairspy import AllPairs
from collections import defaultdict

import os
import itertools
import random
import ase.io


from .iface import BaseCllContext, ICllExploreOutput
from .constant import (
    LAMMPS_DUMP_DIR,
    LAMMPS_DUMP_SUFFIX,
    PRESET_LAMMPS_INPUT_TEMPLATE,
)
from .data import DataFormat, artifacts_to_ase_atoms

logger = get_logger(__name__)


class CllLammpsInputConfig(BaseModel):
    n_wise: int = 0
    """
    The way of combining variants.
    if n_wise is less than 2 or greater than total fields, the full combination will be used.
    Or else, the n_wise combination will be used.
    It is strongly recommended to use n_wise when the full combination is too large.
    """

    explore_vars: Mapping[str, List[Any]]
    """
    Variants to be explored.
    Variables defined here will become **LAMMPS variables**.
    If multiple value has been set for a variable,
    the cartesian product will be used to generate the combination.
    For example,

    ```yaml
    TEMP: [330, 430, 530]  # Can be a scalar, e.g. 330
    PRES: 1                # Can be a vector, e.g. [1, 2, 3]
    LAMBDA_f: [0.0, 0.25, 0.5, 0.75, 1.0]
    ```
    Then you can reference them in the LAMMPS input template as ${TEMP}, ${LAMBDA_f}, ${N_STEPS}, etc.
    """
    preset_template: Optional[str] = None
    """
    Name of the preset template.
    """
    input_template: Optional[str] = None
    """
    LAMMPS input template file content.
    If set, the preset_template will be ignored.
    """
    template_vars: Mapping[str, Any] = dict()
    """
    input_template may provide extra injection points for user to inject custom settings.
    Those value could be set here.
    """

    plumed_config: Optional[str]
    """Plumed config file content."""

    system_files: List[str]
    """
    Artifacts key of lammps input data
    """
    ensemble: Literal['nvt', 'nvt-i', 'nvt-a', 'nvt-iso', 'nvt-aniso', 'npt', 'npt-t', 'npt-tri', 'nve', 'csvr']
    no_pbc: bool = False
    nsteps: int
    timestep: float = 0.0005
    sample_freq: int = 100
    mode: Literal['default', 'fep'] = 'default'


    type_alias: Mapping[str, List[str]] = dict()
    '''
    Type alias for atoms. For example, if you want to distinguish ghost H and H of HF molecule from other H atoms,
    you can define the alias as follows:
    ```yaml
    type_alias:
        H: [ H_ghost, H_hf ]
    ```
    And then you can reference them in the LAMMPS input template, for example
    ```
    set atom 1 type ${H_hf}
    set atom 2 type ${H_ghost}
    ```
    '''

    @validator('explore_vars', pre=True)
    @classmethod
    def validate_explore_variants(cls, value):
        if not isinstance(value, dict):
            raise ValueError('explore_vars must be a dict')
        for k in ['TEMP', 'PRES']:
            if k not in value:
                raise ValueError(f'{k} must be set in explore_variants')
        # override default values
        value = {
            'TAU_T': 0.1,
            'TAU_P': 0.5,
            'TIME_CONST': 0.1,
            **value,
        }
        result = {}
        for k, v in value.items():
            if not isinstance(v, list):
                v = [v]
            result[k] = v
        return result

    @root_validator()
    @classmethod
    def validate_domain(cls, values):
        ensemble = values.get('ensemble')
        no_pbc = values.get('no_pbc')
        if ensemble.startswith('npt') and no_pbc:
            raise ValueError('ensemble npt conflict with no_pcb')
        if not ensemble.startswith('npt'):
            logger.info('ensemble is not npt, force PRES to -1')
            values['explore_vars']['PRES'] = [-1]
        return values


class CllLammpsContextConfig(BaseModel):
    script_template: BashTemplate
    lammps_cmd: str = 'lmp'
    concurrency: int = 5


@dataclass
class CllLammpsInput:
    config: CllLammpsInputConfig
    type_map: List[str]
    mass_map: List[float]
    dp_models: Mapping[str, List[Artifact]]
    preset_template: str


@dataclass
class CllLammpsContext(BaseCllContext):
    config: CllLammpsContextConfig


@dataclass
class GenericLammpsOutput(ICllExploreOutput):
    model_devi_outputs: List[Artifact]

    def get_model_devi_dataset(self) -> List[Artifact]:
        return self.model_devi_outputs


async def cll_lammps(input: CllLammpsInput, ctx: CllLammpsContext):
    executor = ctx.resource_manager.default_executor
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    data_files = ctx.resource_manager.resolve_artifacts(input.config.system_files)
    assert len(data_files) > 0, 'no data files found'

    tasks_dir, task_dirs = executor.run_python_fn(make_lammps_task_dirs)(
        combination_vars=input.config.explore_vars,
        data_files=[a.to_dict() for a in data_files],
        dp_models={k: [m.url for m in v] for k, v in input.dp_models.items()},
        n_steps=input.config.nsteps,
        timestep=input.config.timestep,
        sample_freq=input.config.sample_freq,
        no_pbc=input.config.no_pbc,
        n_wise=input.config.n_wise,
        ensemble=input.config.ensemble,
        preset_template=input.config.preset_template or input.preset_template,
        input_template=input.config.input_template,
        plumed_config=input.config.plumed_config,
        extra_template_vars=input.config.template_vars,
        type_map=input.type_map,
        mass_map=input.mass_map,
        type_alias=input.config.type_alias,
        work_dir=work_dir,
    )

    # build scripts and submit
    base_cmd = f'{ctx.config.lammps_cmd} -i lammps.input'
    cmd = f'''if [ -f md.restart.* ]; then {base_cmd} -v restart 1; else {base_cmd} -v restart 0; fi'''

    # generate steps
    steps = []
    for task_dir in task_dirs:
        steps.append(BashStep(cwd=task_dir['url'], cmd=cmd, checkpoint='lammps'))

    # submit jobs by the number of concurrency
    jobs = []
    for i, steps_group in enumerate(list_split(steps, ctx.config.concurrency)):
        if not steps_group:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps_group,
        )
        job = executor.submit(script.render(), cwd=tasks_dir,
                              checkpoint_key=f'queue-job/lammps/{i}:{tasks_dir}')
        jobs.append(job)

    await gather_jobs(jobs, max_tries=2)

    # build outputs
    outputs = []
    for task_dir in task_dirs:
        common = dict(url=task_dir['url'], executor=executor.name, format=DataFormat.LAMMPS_OUTPUT_DIR)
        # TODO: a more generic way to dealing multiple model_devi outputs
        if input.config.mode == 'fep':
            outputs += [
                Artifact.of(**common, attrs={
                    **task_dir['attrs'], 'model_devi_file': 'model_devi_ini.out', 'lammps_dump_dir': 'traj-ini',
                    **task_dir['attrs']['fep-ini'],
                }),
                Artifact.of(**common, attrs={
                    **task_dir['attrs'], 'model_devi_file': 'model_devi_fin.out', 'lammps_dump_dir': 'traj-fin',
                    **task_dir['attrs']['fep-fin'],
                    'ancestor': task_dir['attrs']['ancestor'] + '-fin',  # only fin needs
                }),
            ]
        else:
            outputs += [
                Artifact.of(**common, attrs={ **task_dir['attrs'] }),
            ]

    return GenericLammpsOutput(model_devi_outputs=outputs)


def __export_remote_functions():

    class LammpsInputTemplate(Template):
        delimiter = '$$'

    def make_lammps_task_dirs(combination_vars: Mapping[str, Sequence[Any]],
                              data_files: List[ArtifactDict],
                              dp_models: Mapping[str, List[str]],
                              n_steps: int,
                              timestep: float,
                              sample_freq: float,
                              no_pbc: bool,
                              n_wise: int,
                              ensemble: str,
                              preset_template: str,
                              input_template: Optional[str],
                              plumed_config: Optional[str],
                              extra_template_vars: Mapping[str, Any],
                              type_map: List[str],
                              mass_map: List[float],
                              type_alias: Mapping[str, List[str]],
                              work_dir: str,
                              ):
        # setup workspace
        input_data_dir = os.path.join(work_dir, 'input_data')
        tasks_dir = os.path.join(work_dir, 'tasks')
        for path in (input_data_dir, tasks_dir):
            os.makedirs(path, exist_ok=True)

        # create data files
        input_dataset = []
        atoms_list = artifacts_to_ase_atoms(data_files, type_map=type_map)
        for i, (artifact, atoms) in enumerate(atoms_list):
            #  create data file
            data_file = os.path.join(input_data_dir, f'{i:06d}.lammps.data')
            ase.io.write(data_file, atoms, format='lammps-data', specorder=type_map)  # type: ignore
            input_dataset.append({
                'url': data_file,
                'attrs': artifact['attrs'],
            })

        # generate combinations of variants
        combination_fields: List[str] = ['DATA_FILE']
        combination_values: List[List[Any]] = [input_dataset]
        for k, v in combination_vars.items():
            combination_fields.append(k)
            combination_values.append(v)  # type: ignore
        if 1 < n_wise <= len(combination_fields):
            logger.info(f'using {n_wise}-wise combination')
            combinations = AllPairs(combination_values, n=n_wise)
        else:
            logger.info('using full combination')
            combinations = itertools.product(*combination_values)

        # generate tasks input
        task_dirs = []
        for i, combination in enumerate(combinations):
            template_vars = {
                'AI2KIT_CMD': 'ai2-kit',  # TODO: this should be configurable via ctx.config instead of template vars
            }
            lammps_vars = dict(zip(combination_fields, combination))

            # setup task dir
            task_dir = os.path.join(tasks_dir, f'{i:06d}')
            os.makedirs(os.path.join(task_dir, LAMMPS_DUMP_DIR), exist_ok=True)

            data_file = lammps_vars.pop('DATA_FILE')

            # override default values with data file attrs
            overridable_params: dict = dict_nested_get(data_file, ['attrs', 'lammps'], dict())  # type: ignore
            plumed_config = overridable_params.get('plumed_config', plumed_config)
            type_alias = overridable_params.get('type_alias', type_alias)
            extra_template_vars = {**extra_template_vars, **overridable_params.get('template_vars', dict())}

            # build type map and type order
            type_to_mass = dict(zip(type_map, mass_map))

            ext_type_map = []
            ext_type_map_to_origin = []
            ext_mass_map = []
            ghost_loc = []  # location of ghost atom type
            DP_GHOST = len(type_map)

            for origin_type, alias in type_alias.items():
                for t in alias:
                    if 'ghost' in t:  # atom type with 'ghost' in its name is considered as ghost atom type
                        ghost_loc.append(DP_GHOST)

                    ext_type_map.append(t)
                    ext_type_map_to_origin.append(origin_type)
                    ext_mass_map.append(type_to_mass[origin_type])

            # SPECORDER is used to specify the order of types in the lammps data file
            # For example, if the complete type_map is [H, O, O_1, O_2, H_1, H_2],
            # then the specorder should be [H, O, O, O, H, H]
            specorder = type_map + ext_type_map_to_origin

            # *_type_order is use to remap the lammps atom type to dp atom type
            # For example, if the full_type_map is     [H, O, O_1, O_2, H_1, H_2],
            # then the fep_ini_type_order should be    [0, 1, 1  , 1,   0  , 0]
            fep_ini_type_order = [ type_map.index(t) for t in specorder]
            # fep_fin_type_order is the same as fep_ini_type_order, except that the ghost atom type should be len(type_map)
            fep_fin_type_order = fep_ini_type_order.copy()
            for loc in ghost_loc:
                fep_fin_type_order[loc] = DP_GHOST

            template_vars['SPECORDER'] = specorder  # type: ignore
            template_vars['SPECORDER_BASE'] = type_map  # type: ignore

            template_vars['DP_DEFAULT_TYPE_ORDER'] = ' '.join(map(str, range(len(type_map))))
            template_vars['DP_FEP_INI_TYPE_ORDER'] = ' '.join(map(str, fep_ini_type_order))
            template_vars['DP_FEP_FIN_TYPE_ORDER'] = ' '.join(map(str, fep_fin_type_order))

            template_vars['MASS_MAP_FULL'] = _get_masses(type_map + ext_type_map, mass_map + ext_mass_map)
            template_vars['MASS_MAP'] =  template_vars['MASS_MAP_FULL']
            template_vars['MASS_MAP_BASE'] = _get_masses(type_map, mass_map)

            ## build variables section
            lammps_vars['DATA_FILE'] = data_file['url']
            lammps_vars['N_STEPS'] = n_steps
            lammps_vars['THERMO_FREQ'] = sample_freq
            lammps_vars['DUMP_FREQ'] = sample_freq
            lammps_vars['SAMPLE_FREQ'] = sample_freq

            dump_json(lammps_vars, os.path.join(task_dir, 'debug.lammps_vars.json'))  # for debug
            template_vars['VARIABLES'] = _get_lammps_variables(lammps_vars)
            ## build init settings
            template_vars['INITIALIZE'] =  '\n'.join([
                'units           metal',
                'atom_style      atomic',
                'boundary ' + ('f f f' if no_pbc else 'p p p'),
            ])
            ## build read data section
            template_vars['READ_DATA'] = (
                '''if "${restart} > 0" '''
                '''then "read_restart md.restart.*" '''
                '''else "read_data ${DATA_FILE} extra/atom/types %s"''' % (len(ext_type_map))
            )

            ## build simulation
            simulation = [
                '''if "${restart} == 0" then "velocity all create ${TEMP} %d"''' % (random.randrange(10^6 - 1) + 1),
                _get_ensemble(ensemble),
            ]

            if plumed_config:
                plumed_config_file = os.path.join(task_dir, 'plumed.input')
                dump_text(plumed_config, plumed_config_file)
                simulation.append(f'fix cll_plumed all plumed plumedfile {plumed_config_file} outfile plumed.out')

            if no_pbc:
                simulation.extend([
                    'velocity all zero linear',
                    'fix      fm all momentum 1 linear 1 1 1',
                ])
            simulation.extend([
                'thermo_style custom step temp pe ke etotal press vol lx ly lz xy xz yz',
                'thermo       ${THERMO_FREQ}',
                'dump         1 all custom ${DUMP_FREQ} %s/*%s id type x y z fx fy fz' % (LAMMPS_DUMP_DIR, LAMMPS_DUMP_SUFFIX),
                'restart      10000 md.restart',
            ])
            template_vars['SIMULATION'] = '\n'.join(simulation)
            ## build run section
            template_vars['RUN'] = '\n'.join([
                'timestep %f' % timestep,
                'run      ${N_STEPS} upto',
            ])

            dp_models_vars = _get_dp_models_variables(dp_models)

            template_vars = {**template_vars, **dp_models_vars, **extra_template_vars}
            dump_json(template_vars, os.path.join(task_dir, 'debug.template_vars.json'))  # for debug

            input_template = PRESET_LAMMPS_INPUT_TEMPLATE[preset_template] if input_template is None else input_template
            dump_text(input_template, os.path.join(task_dir, 'debug.input_template.txt'))  # for debug
            lammps_input = LammpsInputTemplate(input_template).substitute(defaultdict(str),**template_vars)
            dump_text(lammps_input, os.path.join(task_dir, 'lammps.input'))

            task_dirs.append({'url': task_dir, 'attrs': data_file['attrs']})  # type: ignore
        return tasks_dir, task_dirs


    def _get_type_map_vars(type_map: List[str]):
        return dict(zip(type_map, range(1, len(type_map) + 1)))


    def _get_masses(type_map: List[str], mass_map: List[float]):
        lines = [
            _get_lammps_variables(_get_type_map_vars(type_map)),
            '',
        ]
        for t, m in zip(type_map, mass_map):
            lines.append(f'mass ${{{t}}} {m}')
        return '\n'.join(lines)


    def _get_dp_models_variables(models: Mapping[str, List[str]]):
        vars = {}
        for k, v in models.items():
            prefix = 'DP_MODELS' if k == '' else f'DP_{k}_MODELS'
            vars[prefix] = ' '.join(v)
            for i, m in enumerate(v):
                vars[f'{prefix}_{i}'] = m
        return vars


    def _get_lammps_variables(vars: Mapping[str, Any]):
        lines = []
        for k, v in vars.items():
            if isinstance(v, str):
                # TODO: should escape `v` in case of special characters
                line = f'variable    {k:16} string "{v}"'
            else:
                line = f'variable    {k:16} equal {v}'
            lines.append(line)
        return '\n'.join(lines)


    def _get_ensemble(ensemble: str):
        lines = []
        if ensemble in ('npt', 'npt-i', 'npt-iso',):
            lines.append('fix 1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}')
        elif ensemble in ('npt-a', 'npt-aniso',):
            lines.append('fix 1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P}')
        elif ensemble in ('npt-t', 'npt-tri',):
            lines.append('fix 1 all npt temp ${TEMP} ${TEMP} ${TAU_T} tri ${PRES} ${PRES} ${TAU_P}')
        elif ensemble in ('nvt',):
            lines.append('fix 1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}')
        elif ensemble in ('nve',):
            lines.append('fix 1 all nve')
        elif ensemble in ('csvr',):
            lines.append('fix 1 all nve')
            lines.append('fix 2 all temp/csvr ${TEMP} ${TEMP} ${TIME_CONST} %d' % (random.randrange(10^6 - 1) + 1))
        else:
            raise ValueError('unknown ensemble: ' + ensemble)
        return '\n'.join(lines)

    return (
        LammpsInputTemplate,
        make_lammps_task_dirs,
    )


(
    LammpsInputTemplate,
    make_lammps_task_dirs,
) = __export_remote_functions()


import logging

# format to include timestamp and module
logging.basicConfig(format='%(asctime)s %(name)s: %(message)s', level=logging.INFO)
# suppress transitions logging
logging.getLogger('transitions.core').setLevel(logging.WARNING)

def get_logger(name=None):
    return logging.getLogger(name)


from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.script import BashScript, BashStep, BashTemplate
from ai2_kit.core.job import gather_jobs
from ai2_kit.core.util import merge_dict, dict_nested_get, list_split, list_sample, dump_json, dump_text
from ai2_kit.core.log import get_logger

from typing import List, Union, Tuple, Literal, Optional
from pydantic import BaseModel
from dataclasses import dataclass
from ase import Atoms

import copy
import os

from .data import DataFormat, ase_atoms_to_cp2k_input_data, artifacts_to_ase_atoms
from .iface import ICllLabelOutput, BaseCllContext
from .util import loads_cp2k_input, load_cp2k_input, dump_cp2k_input

logger = get_logger(__name__)


class CllCp2kInputConfig(BaseModel):
    init_system_files: List[str] = []
    wfn_warmup_template: Union[None, dict, str] = None
    """
    Warmup template for cp2k. Could be a dict or content of a cp2k input file.
    This template will be used to generate input files for warmup runs.
    The warmup runs can be used to generate wave function files for the main runs.
    """
    input_template: Union[dict, str] = dict()
    """
    Input template for cp2k. Could be a dict or content of a cp2k input file.
    """
    limit: int = 50
    """
    Limit of the number of systems to be labeled.
    """
    limit_method: Literal["even", "random", "truncate"] = "even"


class CllCp2kContextConfig(BaseModel):
    script_template: BashTemplate
    cp2k_cmd: str = 'cp2k'
    post_cp2k_cmd: str = 'echo "no post_cp2k_cmd"'
    concurrency: int = 5


@dataclass
class CllCp2kInput:
    config: CllCp2kInputConfig
    system_files: List[Artifact]
    type_map: List[str]
    initiated: bool = False  # FIXME: this seems to be a bad design idea


@dataclass
class CllCp2kContext(BaseCllContext):
    config: CllCp2kContextConfig


@dataclass
class GenericCp2kOutput(ICllLabelOutput):
    cp2k_outputs: List[Artifact]

    def get_labeled_system_dataset(self):
        return self.cp2k_outputs


async def cll_cp2k(input: CllCp2kInput, ctx: CllCp2kContext) -> GenericCp2kOutput:
    executor = ctx.resource_manager.default_executor

    # For the first round
    # FIXME: move out from this function, this should be done in the workflow
    if not input.initiated:
        input.system_files += ctx.resource_manager.resolve_artifacts(input.config.init_system_files)

    if len(input.system_files) == 0:
        return GenericCp2kOutput(cp2k_outputs=[])

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    [tasks_dir] = executor.setup_workspace(work_dir, ['tasks'])

    # prepare input template
    if isinstance(input.config.input_template, str):
        input_template = loads_cp2k_input(input.config.input_template)
    else:
        input_template = copy.deepcopy(input.config.input_template)

    # create task dirs and prepare input files
    cp2k_task_dirs = executor.run_python_fn(make_cp2k_task_dirs)(
        system_files=[a.to_dict() for a in input.system_files],
        type_map=input.type_map,
        base_dir=tasks_dir,
        input_template=input_template,
        # initialize all data if not initiated
        limit=0 if not input.initiated else input.config.limit,
        wfn_warmup_template=input.config.wfn_warmup_template,
    )

    # build commands
    steps = []
    for cp2k_task_dir in cp2k_task_dirs:
        # run warmup if needed
        # note: use if-else instead of boolean shortcut to avoid wrong status
        cmd = '\n'.join([
            f'if [ -f wfn_warmup.inp ]; then {ctx.config.cp2k_cmd} -i wfn_warmup.inp &> wfn_warmup.out; fi && \\',
            f'{ctx.config.cp2k_cmd} -i input.inp &> output && {ctx.config.post_cp2k_cmd}',
        ])
        steps.append(BashStep(
            cwd=cp2k_task_dir['url'],
            cmd=cmd,
            checkpoint='cp2k',
        ))

    # submit tasks and wait for completion
    jobs = []
    for i, steps_group in enumerate(list_split(steps, ctx.config.concurrency)):
        if not steps_group:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps_group,
        )
        job = executor.submit(script.render(), cwd=tasks_dir,
                              checkpoint_key=f'queue-job:cp2k:{tasks_dir}:{i}')
        jobs.append(job)
    jobs = await gather_jobs(jobs, max_tries=2)

    cp2k_outputs = [Artifact.of(
        url=a['url'],
        format=DataFormat.CP2K_OUTPUT_DIR,
        executor=executor.name,
        attrs=a['attrs'],
    ) for a in cp2k_task_dirs]

    return GenericCp2kOutput(cp2k_outputs=cp2k_outputs)


def __export_remote_functions():

    def make_cp2k_task_dirs(system_files: List[ArtifactDict],
                            type_map: List[str],
                            input_template: dict,
                            base_dir: str,
                            limit: int = 0,
                            sample_method: Literal["even", "random", "truncate"] = "even",
                            input_file_name: str = 'input.inp',
                            wfn_warmup_template: Union[None, dict, str] = None,
                            warmup_file_name: str = 'wfn_warmup.inp'
                            ) -> List[ArtifactDict]:
        """Generate CP2K input files from LAMMPS dump files or XYZ files."""
        task_dirs = []
        atoms_list: List[Tuple[ArtifactDict, Atoms]] = artifacts_to_ase_atoms(system_files, type_map=type_map)

        if limit > 0:
            atoms_list = list_sample(atoms_list, limit, method=sample_method)

        for i, (data_file, atoms) in enumerate(atoms_list):
            # create task dir
            task_dir = os.path.join(base_dir, f'{str(i).zfill(6)}')
            os.makedirs(task_dir, exist_ok=True)
            dump_json(data_file, os.path.join(task_dir, 'debug.data-file.json'))

            # load system-wise config from attrs
            overridable_params: dict = copy.deepcopy(dict_nested_get(data_file, ['attrs', 'cp2k'], dict()))  # type: ignore

            warmup_input = copy.deepcopy(overridable_params.get('wfn_warmup_template') or wfn_warmup_template)
            normal_input = copy.deepcopy(overridable_params.get('input_template') or input_template)

            if warmup_input:
                with open(os.path.join(task_dir, warmup_file_name), 'w') as f:
                    make_cp2k_input(f, warmup_input, atoms)

            with open(os.path.join(task_dir, input_file_name), 'w') as f:
                make_cp2k_input(f, normal_input, atoms)

            task_dirs.append({
                'url': task_dir,
                'attrs': data_file['attrs'],
            })

        return task_dirs


    def make_cp2k_input(fp, input_template: Union[str, dict], atoms: Atoms):
        if isinstance(input_template, str):
            input_template = loads_cp2k_input(input_template)
        assert isinstance(input_template, dict) and input_template, 'input_data must be a non-empty dict'
        coords, cell = ase_atoms_to_cp2k_input_data(atoms)
        merge_dict(input_template, {
            'FORCE_EVAL': {
                'SUBSYS': {
                    # FIXME: this is a dirty hack, we should make dump_cp2k_input support COORD
                    'COORD': dict.fromkeys(coords, ''),
                    'CELL': {
                        'A': ' '.join(map(str, cell[0])),
                        'B': ' '.join(map(str, cell[1])),
                        'C': ' '.join(map(str, cell[2])),
                    }
                }
            }
        })
        dump_cp2k_input(input_template, fp)

    return (
        make_cp2k_task_dirs,
    )


(
    make_cp2k_task_dirs,
) = __export_remote_functions()



from ai2_kit.core.util import ensure_dir

import dpdata
import glob

class DpdataHelper:

    def __init__(self, label: bool = True):
        self._systems = []
        self._label = label

    def read(self, *file_path_or_glob: str,**kwargs):
        kwargs.setdefault('fmt', 'deepmd/npy')
        files = []
        for file_path in file_path_or_glob:
            files += glob.glob(file_path, recursive=True) if '*' in file_path else [file_path]

        if len(files) == 0:
            raise FileNotFoundError(f'No file found for {file_path_or_glob}')
        for file in files:
            self._read(file, **kwargs)
        return self

    def merge_write(self, out_path: str, fmt='deepmd/npy'):
        ensure_dir(out_path)
        if len(self._systems) == 0:
            raise ValueError('No data to merge')

        multi_systems = dpdata.MultiSystems(self._systems[0])
        for system in self._systems[1:]:
            multi_systems.append(system)

        if fmt == 'deepmd/npy':
            multi_systems.to_deepmd_npy(out_path)  # type: ignore
        elif fmt == 'deepmd/raw':
            multi_systems.to_deepmd_raw(out_path)  # type: ignore
        elif fmt == 'deepmd/hdf5':
            multi_systems.to_deepmd_hdf5(out_path)  # type: ignore
        else:
            raise ValueError(f'Unknown fmt {fmt}')

    def _read(self, file: str, **kwargs):
        if self._label:
            self._systems.append(dpdata.LabeledSystem(file, **kwargs))
        else:
            self._systems.append(dpdata.System(file, **kwargs))



from asaplib.data.xyz import ASAPXYZ
from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.log import get_logger
from ai2_kit.core.util import flat_evenly, dump_json, flush_stdio

from typing import List, Optional, Tuple
from io import StringIO
from pydantic import BaseModel
from dataclasses import dataclass
import pandas as pd
from tabulate import tabulate
from itertools import groupby
from operator import itemgetter

import ase.io
import os
import numpy as np

from .data import get_data_format, DataFormat, artifacts_to_ase_atoms
from .iface import ICllSelectorOutput, BaseCllContext
from .constant import LAMMPS_DUMP_DIR, LAMMPS_DUMP_SUFFIX, DEFAULT_ASAP_SOAP_DESC, DEFAULT_ASAP_PCA_REDUCER
from .asap import get_descriptor, reduce_dimension, get_trainer, get_cluster


logger = get_logger(__name__)


class CllModelDeviSelectorInputConfig(BaseModel):
    class AsapOptions(BaseModel):
        disable: bool = False
        max_structures_per_system: int = 10
        """
        max number of structures to be selected from the same group
        """
        descriptor: dict = {'soap': { **DEFAULT_ASAP_SOAP_DESC, 'preset': 'minimal'}}
        dim_reducer: dict = {'pca': DEFAULT_ASAP_PCA_REDUCER}
        cluster: dict = {'dbscan': {}}

    f_trust_lo: float
    f_trust_hi: float
    asap_options: Optional[AsapOptions]


@dataclass
class CllModelDevSelectorContext(BaseCllContext):
    ...


@dataclass
class CllModelDeviSelectorOutput(ICllSelectorOutput):
    candidates: List[Artifact]
    passing_rate: float

    def get_model_devi_dataset(self):
        return self.candidates

    def get_passing_rate(self) -> float:
        return self.passing_rate


@dataclass
class CllModelDeviSelectorInput:
    config: CllModelDeviSelectorInputConfig
    model_devi_data: List[Artifact]
    model_devi_file: str
    type_map: List[str]

    def set_model_devi_dataset(self, data: List[Artifact]):
        self.model_devi_data = data


async def cll_model_devi_selector(input: CllModelDeviSelectorInput, ctx: CllModelDevSelectorContext):
    executor = ctx.resource_manager.default_executor
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    executor.mkdir(work_dir)

    f_trust_lo = input.config.f_trust_lo
    f_trust_hi = input.config.f_trust_hi

    results = executor.run_python_fn(bulk_select_structures_by_model_devi)(
        model_devi_outputs=[a.to_dict() for a in input.model_devi_data],
        model_devi_file=input.model_devi_file,
        f_trust_lo=f_trust_lo, f_trust_hi=f_trust_hi,
        type_map=input.type_map, work_dir=work_dir,
    )

    candidates = [candidate for candidate, _ in results if candidate is not None]  # filter out None
    stats = [stats for _, stats in results]

    # write model_deviation stats report
    def _get_row(stat: dict):
        url = stat['src']
        url = f'...{url[-30:]}' if len(url) > 30 else url  # wrap url if it is too long
        total, good, decent, poor = stat['total'], stat['good'], stat['decent'], stat['poor']
        return [
            url, total, good, decent, poor,
            f'{good / total * 100:.2f}%', f'{decent / total * 100:.2f}%', f'{poor / total * 100:.2f}%',
        ]
    headers = ['file', 'total', 'good', 'decent', 'poor', 'good%', 'decent%', 'poor%']
    table = [ _get_row(stat) for stat in stats]
    total = sum(row[1] for row in table)
    total_good = sum(row[2] for row in table)

    stats_report = tabulate(table, headers=headers, tablefmt='tsv')
    logger.info('stats report: \n%s\n', stats_report)
    executor.dump_text(stats_report, os.path.join(work_dir, 'stats.tsv'))

    # further select candidates by ASAP
    if input.config.asap_options and not input.config.asap_options.disable:
        asap_options = input.config.asap_options
        candidates = executor.run_python_fn(bulk_select_distinct_structures)(
            candidates=candidates,
            descriptor_opt=asap_options.descriptor,
            dim_reducer_opt=asap_options.dim_reducer,
            cluster_opt=asap_options.cluster,
            type_map=input.type_map,
            work_dir=work_dir,
            max_structures_per_system=asap_options.max_structures_per_system,
        )

    return CllModelDeviSelectorOutput(
        candidates=[ Artifact.of(**a) for a in candidates ],
        passing_rate=total_good / total,
    )


def __export_remote_functions():

    def bulk_select_structures_by_model_devi(model_devi_outputs: List[ArtifactDict],
                                             model_devi_file: str,
                                             f_trust_lo: float,
                                             f_trust_hi: float,
                                             type_map: List[str],
                                             work_dir: str,
                                             workers: int = 4,
                                             ) -> List[Tuple[Optional[ArtifactDict], dict]]:
        import joblib
        return joblib.Parallel(n_jobs=workers)(
            joblib.delayed(select_structures_by_model_devi)(
                model_devi_output=output,
                model_devi_file=model_devi_file,
                f_trust_lo=f_trust_lo, f_trust_hi=f_trust_hi,
                type_map=type_map,
                work_dir=os.path.join(work_dir, 'model_devi', f'{i:06}')
            )
            for i, output in enumerate(model_devi_outputs)
        )  # type: ignore


    def select_structures_by_model_devi(model_devi_output: ArtifactDict,
                                        model_devi_file: str,
                                        f_trust_lo: float,
                                        f_trust_hi: float,
                                        type_map: List[str],
                                        work_dir: str,
                                        ) -> Tuple[Optional[ArtifactDict], dict]:
        """
        analysis the model_devi output of explore stage and select candidates
        """


        os.makedirs(work_dir, exist_ok=True)
        dump_json(model_devi_output, os.path.join(work_dir, 'model_devi_output.debug.json'))

        model_devi_dir = model_devi_output['url']
        model_devi_file = model_devi_output['attrs'].pop('model_devi_file', model_devi_file)

        force_col = 'max_devi_f'
        print(f'criteria: {f_trust_lo} <= {force_col} < {f_trust_hi}')

        # get path of model_devi file
        data_format = get_data_format(model_devi_output)  # type: ignore
        if data_format in (DataFormat.LAMMPS_OUTPUT_DIR, DataFormat.LASP_LAMMPS_OUT_DIR):
            model_devi_file = os.path.join(model_devi_dir, model_devi_file)
        else:
            raise ValueError('unknown model_devi_data types')
        logger.info('start to analysis file: %s', model_devi_file)

        # load model_devi data
        with open(model_devi_file, 'r') as f:
            text = f.read()
        df = pd.read_csv(StringIO(text.lstrip('#')), delim_whitespace=True)
        # layout:
        #        step  max_devi_v  min_devi_v  avg_devi_v  max_devi_f  min_devi_f  avg_devi_f
        # 0        0    0.006793    0.000672    0.003490    0.143317    0.005612    0.026106
        # 1      100    0.006987    0.000550    0.003952    0.128178    0.006042    0.022608

        # evaluate new found structures by their model_devi score in 3 levels: good, decent, poor
        good_df   = df[df[force_col] < f_trust_lo]
        decent_df = df[(df[force_col] >= f_trust_lo) & (df[force_col] < f_trust_hi)]
        poor_df   = df[df[force_col] >= f_trust_hi]

        stats = {
            'src': model_devi_file,
            'total': len(df),
            'good': len(good_df),
            'decent': len(decent_df),
            'poor': len(poor_df),
        }

        # load structures
        atoms_list = []
        if data_format == DataFormat.LAMMPS_OUTPUT_DIR:
            lammps_dump_dir = model_devi_output['attrs'].pop('lammps_dump_dir', LAMMPS_DUMP_DIR)
            for frame_id in df.step:
                dump_file = os.path.join(model_devi_dir, lammps_dump_dir, f'{frame_id}{LAMMPS_DUMP_SUFFIX}')
                atoms_list += ase.io.read(dump_file, ':', format='lammps-dump-text', specorder=type_map)
        elif data_format == DataFormat.LASP_LAMMPS_OUT_DIR:
            structures_file = os.path.join(model_devi_dir, 'structures.xyz')
            atoms_list += ase.io.read(structures_file, ':', format='extxyz')
        else:
            raise ValueError('unknown model_devi_data types')

        output_artifact = None
        # dump structures to different files
        if len(good_df) > 0:
            good_file = os.path.join(work_dir, 'good.xyz')
            ase.io.write(good_file, list(itemgetter(*good_df.index)(atoms_list)), format='extxyz')
        if len(poor_df) > 0:
            poor_file = os.path.join(work_dir, 'poor.xyz')
            ase.io.write(poor_file, list(itemgetter(*poor_df.index)(atoms_list)), format='extxyz')
        if len(decent_df) > 0:
            decent_file = os.path.join(work_dir, 'decent.xyz')
            ase.io.write(decent_file, list(itemgetter(*decent_df.index)(atoms_list)), format='extxyz')
            output_artifact = {
                'url': decent_file,
                'format': DataFormat.EXTXYZ,
                'attrs': {
                    **model_devi_output['attrs'],
                }
            }
        dump_json([output_artifact, stats], os.path.join(work_dir, 'output.debug.json'))
        return output_artifact, stats  # type: ignore

    def bulk_select_distinct_structures(candidates: List[ArtifactDict],
                                        descriptor_opt: dict,
                                        dim_reducer_opt: dict,
                                        cluster_opt: dict,
                                        type_map: List[str],
                                        work_dir: str,
                                        max_structures_per_system: int = -1,
                                        workers: int = 4,
                                        ) -> List[ArtifactDict]:
        inputs = []
        for i, (ancestor_key, candidate_group) in enumerate(groupby(candidates, key=lambda c: c['attrs']['ancestor'])):
            candidate_group = list(candidate_group)
            inputs.append((candidate_group, candidate_group[0]['attrs']))

        import joblib
        return joblib.Parallel(n_jobs=workers)(
            joblib.delayed(select_distinct_structures)(
                candidates=group,
                attrs=attrs,
                descriptor_opt=descriptor_opt,
                dim_reducer_opt=dim_reducer_opt,
                cluster_opt=cluster_opt,
                type_map=type_map,
                work_dir=os.path.join(work_dir, 'asap', f'{i:06}'),
                max_structures_per_system=max_structures_per_system,
            ) for i, (group, attrs) in enumerate(inputs)
        )  # type: ignore


    def select_distinct_structures(candidates: List[ArtifactDict],
                                   attrs: dict,
                                   descriptor_opt: dict,
                                   dim_reducer_opt: dict,
                                   cluster_opt: dict,
                                   type_map: List[str],
                                   work_dir: str,
                                   max_structures_per_system: int = -1,
                                   ):
        os.makedirs(work_dir, exist_ok=True)

        dump_json(attrs, os.path.join(work_dir, 'attrs.debug.json'))
        # load structures and save it to a tmp file for ASAP to load
        atoms_list = [atoms for _, atoms in artifacts_to_ase_atoms(candidates, type_map=type_map)]

        if len(atoms_list) < max(max_structures_per_system, 10):
            # Nothing to do when the total number of atoms is small
            # FIXME: there are a lot of potential issue when the number of atoms is small
            # the root cause is in asaplib, which I guess has not been tested with small dataset
            selected_atoms_list = atoms_list
        else:
            tmp_structures_file = os.path.join(work_dir, '.tmp-structures.xyz')
            ase.io.write(tmp_structures_file, atoms_list, format='extxyz')

            # use asaplib to group structures
            asapxyz = ASAPXYZ(tmp_structures_file)

            # group structures
            path_prefix = os.path.join(work_dir, 'asap')
            descriptors, _ = get_descriptor(asapxyz, descriptor_opt, path_prefix=path_prefix)
            reduced_descriptors = reduce_dimension(descriptors, dim_reducer_opt)
            trainer = get_trainer(reduced_descriptors, cluster_opt)
            cluster_labels = get_cluster(asapxyz, reduced_descriptors, trainer, path_prefix=path_prefix)

            # if max_structures_per_system is not set
            # selected at least 1 structure from each group
            if max_structures_per_system <= 0:
                max_structures_per_system = len(cluster_labels)

            # unpack clusters into a list and
            # ensure frames of the same group won't be next to each other
            # so that we can pick up distinct structures by simply choose the first N frames
            order_frames = flat_evenly(cluster_labels.values())
            selected_frames = order_frames[:max_structures_per_system]
            selected_atoms_list = [atoms_list[i] for i in selected_frames]

        # write selected structures to file
        distinct_structures_file = os.path.join(work_dir,  'distinct_structures.xyz')
        ase.io.write(distinct_structures_file, selected_atoms_list, format='extxyz')

        output = {
            'url': distinct_structures_file,
            'format': DataFormat.EXTXYZ,
            'attrs': attrs,
        }
        dump_json(output, os.path.join(work_dir, 'output.debug.json'))
        flush_stdio()  # flush joblib stdio buffer
        return output

    return (
        bulk_select_structures_by_model_devi,
        bulk_select_distinct_structures,
    )

(
    bulk_select_structures_by_model_devi,
    bulk_select_distinct_structures,
) = __export_remote_functions()



from typing import Generic, Optional, TypeVar
from abc import abstractmethod


T = TypeVar('T')


class IFuture(Generic[T]):

    @abstractmethod
    def done(self) -> bool:
        pass

    @abstractmethod
    def result(self, timeout: Optional[float] = None) -> T:
        pass

    async def result_async(self, timeout: Optional[float] = None) -> T:
        ...


from typing import Dict, List, Tuple, Union, Optional, Sequence
import copy

from .artifact import Artifact, ArtifactMap
from .executor import Executor, ExecutorMap, create_executor


ArtifactOrKey = Union[Artifact, str]

class ResourceManager:

    @property
    def default_executor(self):
        return self.get_executor()

    def __init__(self,
                 executor_configs: ExecutorMap,
                 artifacts: ArtifactMap,
                 default_executor: str,
                 ) -> None:

        self._executor_configs = executor_configs
        self._default_executor_name = default_executor
        self._executors: Dict[str, Executor] = dict()
        # runtime check to ensure quick failure
        self.default_executor

        # fill in default values
        for key, artifact in artifacts.items():
            if artifact.executor is None:
                artifact.executor = self.default_executor.name
            if artifact.key is None:
                artifact.key = key
        self._artifacts = artifacts

    def get_executor(self, name: Optional[str] = None) -> Executor:
        if name is None:
            name = self._default_executor_name

        config = self._executor_configs.get(name)
        if config is None:
            raise ValueError(
                'Executor with name {} is not found!'.format(name))

        if name not in self._executors:
            executor = create_executor(config, name)
            executor.init()
            self._executors[name] = executor

        return self._executors[name]

    def get_artifact(self, key: str) -> Artifact:
        # raise error it is by designed, ensure quick failure
        return self._artifacts[key]

    def get_artifacts(self, keys: List[str]) -> List[Artifact]:
        return [self.get_artifact(key) for key in keys]

    def resolve_artifact(self, artifact: ArtifactOrKey ) -> List[Artifact]:
        # TODO: support cross executor data resolve in the future
        if isinstance(artifact, str):
            artifact = self.get_artifact(artifact)
        paths = self.default_executor.resolve_artifact(artifact)

        result = [Artifact.of(
            url=path,
            format=artifact.format,
            includes=None,  # has been consumed
            attrs=copy.deepcopy(artifact.attrs),
            executor=self.default_executor.name,
        ) for path in paths]

        assert len(result) > 0, f'artifact {artifact} is invalid'
        return result

    def resolve_artifacts(self, artifacts: Sequence[ArtifactOrKey]) -> List[Artifact]:
        # flat map
        return [x for a in artifacts for x in self.resolve_artifact(a)]



from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.script import BashScript, BashStep, BashTemplate
from ai2_kit.core.job import gather_jobs
from ai2_kit.core.util import dict_nested_get, list_split, list_sample
from ai2_kit.core.log import get_logger

from typing import List, Union, Literal
from pydantic import BaseModel
from dataclasses import dataclass

from typing import List, Union, Optional, Tuple
from pymatgen.io.vasp.inputs import Incar, Kpoints
from pydantic import BaseModel
from dataclasses import dataclass

import copy
import os

from .data import DataFormat, artifacts_to_ase_atoms
from .iface import ICllLabelOutput, BaseCllContext

logger = get_logger(__name__)

class CllVaspInputConfig(BaseModel):
    init_system_files: List[str] = []
    input_template: Union[dict, str]
    potcar_source: Union[dict, list]
    kpoints_template: Optional[Union[dict, str]] = None
    """
    Input template for VASP. Could be a dict or content of a VASP input file.
    """
    limit: int = 50
    limit_method: Literal["even", "random", "truncate"] = "even"

class CllVaspContextConfig(BaseModel):
    script_template: BashTemplate
    vasp_cmd: str = 'vasp_std'
    concurrency: int = 5

@dataclass
class CllVaspInput:
    config: CllVaspInputConfig
    system_files: List[Artifact]
    type_map: List[str]
    initiated: bool = False


@dataclass
class CllVaspContext(BaseCllContext):
    config: CllVaspContextConfig


@dataclass
class GenericVaspOutput(ICllLabelOutput):
    vasp_outputs: List[Artifact]

    def get_labeled_system_dataset(self):
        return self.vasp_outputs


async def cll_vasp(input: CllVaspInput, ctx: CllVaspContext) -> GenericVaspOutput:
    executor = ctx.resource_manager.default_executor

    # For the first round
    if not input.initiated:
        input.system_files += ctx.resource_manager.resolve_artifacts(input.config.init_system_files)

    if len(input.system_files) == 0:
        return GenericVaspOutput(vasp_outputs=[])

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    [tasks_dir] = executor.setup_workspace(work_dir, ['tasks'])

    # prepare input template
    if isinstance(input.config.input_template, str):
        input_template = input.config.input_template
        input_template = Incar.from_file(input_template).as_dict()
    else:
        input_template = copy.deepcopy(input.config.input_template)

    # prepare potcar
    if isinstance(input.config.potcar_source, dict):
        potcar_source = input.config.potcar_source
    elif isinstance(input.config.potcar_source, list):
        # default: use same sequence as type_map
        if len(input.config.potcar_source) >= len(input.type_map):
            potcar_source = {
                k: v for k, v in zip(input.type_map, input.config.potcar_source)
            }
        else:
            raise ValueError('potcar_source should not be shorter than type_map')
    else:
        # TODO: support generate POTCAR from given path of potential.
        raise ValueError('potcar_source should be either dict or list')

    # prepare kpoints
    kpoints_template = input.config.kpoints_template
    if isinstance(kpoints_template, str):
        kpoints_template = Kpoints.from_file(kpoints_template).as_dict()
    elif isinstance(kpoints_template, dict):
        kpoints_template = copy.deepcopy(kpoints_template)
    else:
        kpoints_template = None

    system_files = ctx.resource_manager.resolve_artifacts(input.system_files)

    # create task dirs and prepare input files
    vasp_task_dirs = executor.run_python_fn(make_vasp_task_dirs)(
        system_files=[a.to_dict() for a in system_files],
        type_map=input.type_map,
        base_dir=tasks_dir,
        input_template=input_template,
        potcar_source=potcar_source,
        kpoints_template=kpoints_template,
        limit=input.config.limit,
    )

    # build commands
    steps = []
    for vasp_task_dir in vasp_task_dirs:
        steps.append(BashStep(
            cwd=vasp_task_dir['url'],
            cmd=[ctx.config.vasp_cmd, ' 1>> output 2>> output'],
            checkpoint='vasp',
        ))

    # submit tasks and wait for completion
    jobs = []
    for i, steps_group in enumerate(list_split(steps, ctx.config.concurrency)):
        if not steps_group:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps_group,
        )
        job = executor.submit(script.render(), cwd=tasks_dir,
                              checkpoint_key=f'queue-job:vasp:{tasks_dir}:{i}')
        jobs.append(job)
    jobs = await gather_jobs(jobs, max_tries=2)

    vasp_outputs = [Artifact.of(
        url=a['url'],
        format=DataFormat.VASP_OUTPUT_DIR,
        executor=executor.name,
        attrs=a['attrs'],
    ) for a in vasp_task_dirs]

    return GenericVaspOutput(vasp_outputs=vasp_outputs)


def __export_remote_functions():
    def make_vasp_task_dirs(system_files: List[ArtifactDict],
                            type_map: List[str],
                            input_template: dict,
                            potcar_source: dict,
                            base_dir: str,
                            kpoints_template: Optional[dict] = None,
                            limit: int = 0,
                            sample_method: Literal["even", "random", "truncate"] = "even"
                            ) -> List[ArtifactDict]:
        """Generate VASP input files from LAMMPS dump files or XYZ files."""

        import ase.io
        from ase import Atoms
        from ase.io.vasp import _symbol_count_from_symbols

        task_dirs = []
        atoms_list: List[Tuple[ArtifactDict, Atoms]] = artifacts_to_ase_atoms(system_files, type_map)

        if limit > 0:
            atoms_list = list_sample(atoms_list, limit, method=sample_method)

        for i, (file, atoms) in enumerate(atoms_list):
            # create task dir
            task_dir = os.path.join(base_dir, f'{str(i).zfill(6)}')
            os.makedirs(task_dir, exist_ok=True)

            # create input file
            input_data = copy.deepcopy(input_template)
            input_data = dict_nested_get(
                file, ['attrs', 'vasp', 'input_data'], input_data # type: ignore
            )
            incar = Incar.from_dict(input_data)
            incar.write_file(os.path.join(task_dir, 'INCAR'))

            # create POSCAR
            elements_all = atoms.get_chemical_symbols()
            elements = [
                item[0] for item in _symbol_count_from_symbols(elements_all)
            ]
            ase.io.write(
                os.path.join(task_dir, 'POSCAR'), atoms, format='vasp'
            )

            # create POTCAR
            with open(os.path.join(task_dir, 'POTCAR'), 'w') as out_file:
                for element in elements:
                    with open(potcar_source[element], 'r') as in_file:
                        out_file.write(in_file.read())

            # create KPOINTS
            kpoints_template = dict_nested_get(
                file, ['attrs', 'vasp', 'kpoints_template'], None # type: ignore
            )
            if kpoints_template:
                kpoints = Kpoints.from_dict(kpoints_template)
                kpoints.write_file(os.path.join(task_dir, 'KPOINTS'))

            # inherit attrs from input file
            # TODO: inherit only ancestor key should be enough
            task_dirs.append({
                'url': task_dir,
                'attrs': file['attrs'],
            })

        return task_dirs

    return make_vasp_task_dirs
make_vasp_task_dirs = __export_remote_functions()



from fire import Fire


class Group:
    def __init__(self, items: dict, doc: str = '') -> None:
        self.__doc__ = doc
        self.__dict__.update(items)


class ProtonTransferGroup:
    @property
    def analyze(self):
        from ai2_kit.algorithm import proton_transfer
        return proton_transfer.proton_transfer_detection

    @property
    def visualize(self):
        from ai2_kit.algorithm import proton_transfer
        return proton_transfer.visualize_transfer

    @property
    def show_transfer_paths(self):
        from ai2_kit.algorithm import proton_transfer
        return proton_transfer.analysis_transfer_paths

    @property
    def show_type_change(self):
        from ai2_kit.algorithm import proton_transfer
        return proton_transfer.detect_type_change

    @property
    def calculate_distances(self):
        from ai2_kit.algorithm import proton_transfer
        return proton_transfer.calculate_distances

    @property
    def show_distance_change(self):
        from ai2_kit.algorithm import proton_transfer
        return proton_transfer.show_distance_change


class WorkflowGroup:
    @property
    def cll_mlp_training(self):
        from ai2_kit.workflow.cll_mlp import run_workflow
        return run_workflow

    @property
    def fep_mlp_training(self):
        from ai2_kit.workflow.fep_mlp import run_workflow
        return run_workflow

class ToolGroup:

    @property
    def ase(self):
        from ai2_kit.tool.ase import AseHelper
        return AseHelper

    @property
    def dpdata(self):
        from ai2_kit.tool.dpdata import DpdataHelper
        return DpdataHelper

    @property
    def checkpoint(self):
        from ai2_kit.core.checkpoint import CheckpointCmd
        return CheckpointCmd


ai2_kit = Group({
    'workflow': WorkflowGroup(),
    'algorithm': Group({
        'proton-transfer': ProtonTransferGroup(),
    }),
    'tool': ToolGroup(),
}, doc="Welcome to use ai2-kit!")


def main():
    Fire(ai2_kit)

if __name__ == '__main__':
    main()
